@@ -9,6 +9,7 @@
 #'
 #' @docType package
 #' @name reticulate
+#' @keywords internal
 #' @useDynLib reticulate, .registration = TRUE
 #' @importFrom Rcpp evalCpp
 NULL
@@ -22,7 +23,16 @@ NULL
 .globals$delay_load_environment <- NULL
 .globals$delay_load_priority <- 0
 .globals$suppress_warnings_handlers <- list()
-.globals$class_filters <- list()
+.globals$class_filters <- list(
+
+  function(classes) {
+    if ("python.builtin.BaseException" %in% classes) {
+      classes <- unique(c(classes, "error", "condition"))
+    }
+    classes
+  }
+
+)
 .globals$py_repl_active <- FALSE
 
 is_python_initialized <- function() {
@@ -32,56 +42,58 @@ is_python_initialized <- function() {
 
 ensure_python_initialized <- function(required_module = NULL) {
 
-  if (!is_python_initialized()) {
+  # nothing to do if python is initialized
+  if (is_python_initialized())
+    return()
 
-    # give delay load modules priority
-    use_environment <- NULL
-    if (!is.null(.globals$delay_load_module)) {
-      required_module <- .globals$delay_load_module
-      use_environment <- .globals$delay_load_environment
-      .globals$delay_load_module <- NULL # one shot
-      .globals$delay_load_environment <- NULL
-      .globals$delay_load_priority <- 0
-    }
-
-    # notify front-end (if any) that Python is about to be initialized
-    callback <- getOption("reticulate.python.beforeInitialized")
-    if (is.function(callback))
-      callback()
-
-    # perform initialization
-    .globals$py_config <- initialize_python(required_module, use_environment)
-
-    # remap output streams to R output handlers
-    remap_output_streams()
-
-    # generate 'R' helper object
-    py_inject_r()
-
-    # inject hooks
-    py_inject_hooks()
-
-    # install required packages
-    configure_environment()
-    
-    # notify front-end (if any) that Python has been initialized
-    callback <- getOption("reticulate.python.afterInitialized")
-    if (is.null(callback))
-      callback <- getOption("reticulate.initialized")
-
-    if (is.function(callback))
-      callback()
-
-    # set up a Python signal handler
-    signals <- import("rpytools.signals")
-    signals$initialize(py_interrupts_pending)
-    
-    # register C-level interrupt handler
-    py_register_interrupt_handler()
-
-    # call init hooks
-    call_init_hooks()
+  # give delay load modules priority
+  use_environment <- NULL
+  if (!is.null(.globals$delay_load_module)) {
+    required_module <- .globals$delay_load_module
+    use_environment <- .globals$delay_load_environment
+    .globals$delay_load_module <- NULL # one shot
+    .globals$delay_load_environment <- NULL
+    .globals$delay_load_priority <- 0
   }
+
+  # notify front-end (if any) that Python is about to be initialized
+  callback <- getOption("reticulate.python.beforeInitialized")
+  if (is.function(callback))
+    callback()
+
+  # perform initialization
+  .globals$py_config <- initialize_python(required_module, use_environment)
+
+  # remap output streams to R output handlers
+  remap_output_streams()
+
+  # generate 'R' helper object
+  py_inject_r()
+
+  # inject hooks
+  py_inject_hooks()
+
+  # install required packages
+  configure_environment()
+
+  # notify front-end (if any) that Python has been initialized
+  callback <- getOption("reticulate.python.afterInitialized")
+  if (is.null(callback))
+    callback <- getOption("reticulate.initialized")
+
+  if (is.function(callback))
+    callback()
+
+  # set up a Python signal handler
+  signals <- import("rpytools.signals")
+  signals$initialize(py_interrupts_pending)
+
+  # register C-level interrupt handler
+  py_register_interrupt_handler()
+
+  # call init hooks
+  call_init_hooks()
+
 }
 
 
@@ -107,12 +119,16 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
     required_module <- strsplit(required_module, ".", fixed = TRUE)[[1]][[1]]
 
   # find configuration
-  config <- py_discover_config(required_module, use_environment)
+  config <- local({
+    op <- options(reticulate.python.initializing = TRUE)
+    on.exit(options(op), add = TRUE)
+    py_discover_config(required_module, use_environment)
+  })
 
   # check if R is embedded in an python environment
   py_embedded <- !is.null(main_process_python_info())
 
-  # check for basic python prerequsities
+  # check for basic python pre-requisites
   if (is.null(config)) {
     python_not_found("Installation of Python not found, Python bindings not loaded.")
   } else if (!is_windows() && is.null(config$libpython)) {
@@ -156,23 +172,21 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
 
   # munge PATH for python (needed so libraries can be found in some cases)
   oldpath <- python_munge_path(config$python)
-  
+
   # on macOS, we need to do some gymnastics to ensure that Anaconda
   # libraries can be properly discovered (and this will only work in RStudio)
   if (is_osx()) local({
-    
+
     symlink <- Sys.getenv("RSTUDIO_FALLBACK_LIBRARY_PATH", unset = NA)
     if (is.na(symlink))
       return()
-    
-    if (file.exists(symlink))
-      unlink(symlink)
-    
+
+    unlink(symlink)
     target <- dirname(config$libpython)
     file.symlink(target, symlink)
-    
+
   })
-  
+
   # initialize python
   tryCatch({
 
@@ -222,18 +236,23 @@ initialize_python <- function(required_module = NULL, use_environment = NULL) {
   config$available <- TRUE
 
   if (py_embedded) {
-    
+
     # we need to insert path to rpytools directly for embedded R
     path <- system.file("python", package = "reticulate")
     fmt <- "import sys; sys.path.append(%s)"
     cmd <- sprintf(fmt, shQuote(path))
-    
+
     py_run_string_impl(cmd)
-    
+
+  }
+
+  if (is_windows()) {
+    # patch sys.executable to point to python.exe, not Rterm.exe or rsession-utf8.exe, #1258
+    py_run_string_impl("import sys; sys.executable = sys.argv[0]", local = TRUE)
   }
 
   # ensure modules can be imported from the current working directory
-  py_run_string_impl("import sys; sys.path.insert(0, '')")
+  py_run_string_impl("import sys; sys.path.insert(0, '')", local = TRUE)
 
   # if this is a conda installation, set QT_QPA_PLATFORM_PLUGIN_PATH
   # https://github.com/rstudio/reticulate/issues/586
