@@ -44,6 +44,14 @@ config = {
   "LIBDIR"           : sysconfig.get_config_var("LIBDIR")
 }
 
+# detect if this is a conda managed python
+# https://stackoverflow.com/a/21282816/5128728
+if sys.version_info >= (3, 7):
+  is_conda = os.path.exists(os.path.join(sys.prefix, 'conda-meta'))
+else:
+  is_conda = 'conda' in sys.version
+config['IsConda'] = is_conda
+
 # Read numpy configuration (if available)
 try:
   import numpy
